% Estimates a good exposure time sequence for sampling a HDR scene
% in a way that a minimum SNR is obtained at every pixel.
%
% The script was used to generate the results reported in the paper:
%
% Optimal HDR reconstruction with linear digital cameras
% M. Granados, B. Ajdin, M. Wand, C. Theobalt, H.-P. Seidel, H. P. A. Lensch
% In Proc. IEEE Conf. Comp. Vis. Pat. Rec. (CVPR), 
%   June 13-18, 2010, San Francisco, USA
%
% Author: Miguel Granados, MPI Informatik, 2011

function recon = dotopt(cam, recon)

recon.t = []; % estimated exposure sequence

X_plot_process = false;
if X_plot_process
    [ts tsi] = sort(recon.X.mu(:));
end

% initialize exposure time limits
recon.X.min = min(recon.X.mu(:));
recon.X.max = max(recon.X.mu(:));
recon.tsnr = 10^(recon.tsnr0/20); % target SNR
recon.tmax = (cam.vsat - (cam.R.mu+12*sqrt(cam.R.var))) / ...
    (cam.g * (recon.X.min + cam.D.mu));
recon.tmin = 12*sqrt(cam.R.var) / (cam.g * (recon.X.max + cam.D.mu));

% initialize exposure time search space
recon.taltn = 50;
recon.tbase = 1.3;
recon.talt = recon.tbase.^linspace(log(recon.tmin)/log(recon.tbase), ...
    log(recon.tmax)/log(recon.tbase), ...
    recon.taltn);

% initialize search space plot
if X_plot_process
    figure(1), clf, hold on
    xlabel('Exposure time (s)');
    ylabel('Weighted sum SNR');
    % initialize snr plot
    figure(2), clf, hold on, %ylim([0 20*log10(recon.tsnr)*1.5]);
    minx = min(recon.X.mu(:));
    maxx = max(recon.X.mu(:));
    xlabel('Irradiance');
    ylabel('SNR (dB)');
    semilogx([minx*.8 maxx*1.2], ...
        [20*log10(recon.tsnr) 20*log10(recon.tsnr)], '-k');
    text(maxx*1.2, 20*log10(recon.tsnr), sprintf('%ddB', recon.tsnr0));
end

% initialize current estimation
muxS = zeros(size(recon.X.mu));
wmuxS = zeros(size(recon.X.mu));
varmuxS = zeros(size(recon.X.mu));

ei = 1;
recon.minsnr = 1;
while (ei <= recon.nt && recon.minsnr < recon.tsnr)
    recon.tscore2 = -Inf(recon.taltn);
    recon.tsnrmin = Inf(recon.taltn);
    
    % compute exposure time cost
    for i=1:recon.taltn
        recon.t(ei) = recon.talt(i);
        [snrmux varmux mux varmuxSi muxSi wmuxSi] = ...
            topt_evalvarx(cam, recon.X.mu, recon.t(1:ei), false, ...
            varmuxS, muxS, wmuxS);
        recon.tsnrmin(i) = min(snrmux(:));
        
        
        for j=1:recon.taltn
            recon.t(ei+1) = recon.talt(j);
            [snrmux varmux mux] = ...
                topt_evalvarx(cam, recon.X.mu, recon.t, false, ...
                varmuxSi, muxSi, wmuxSi);
            if(any(isinf(varmux(:))))
                continue;
            end
            % A1: maximize the SUM of ratios x/sigx
            % Result: all samples at maximum non-saturating exposure time
            % Idea: no saturation means all samples contribute to increase the SNR
            %recon.tscore2(i,j) = sum(snrmux(:));
            
            % A2: maximize the minimum ratio x/sigx
            % Result: increasingly shorter exposure times
            % Idea: ?, should be increasingly larger exposure times
            %recon.tscore2(i,j) = min(snrmux(:));
            
            % A3: minimize the sum of snr reciprocals
            % Result: repeated short exposures mixed with a few larger ones
            % Idea: since there are few low snrs, samples are also given to
            % larger exposures to decrease the large number of high snrs
            %recon.tscore2(i,j) = -sum(1./snrmux(:));
            
            % A4: minimize the sum of exp(snr) reciprocals (SELECTED)
            % Result: distinct larger exposures mixed with a few short ones
            % Idea: larger snr are exponentially reduced so only small ones
            % prevail.
            % Warning1: bottle neck remains on high irradiances, which need
            % many short exposures to improve the snr by just a factor of
            % sqrt(N), which is just adding a constant to 20*log10(r)
            recon.tscore2(i,j) = -log(sum(1./exp(snrmux(:))));
            
            %A6
            %recon.tscore2(i,j) = mean(snrmux(:))/std(snrmux(:));
        end
    end
    
    % select optimal exposure time
    [smin sminij] = max(recon.tscore2(:));
    smini = floor((sminij-1) / recon.taltn)+1;
    sminj = mod((sminij-1), recon.taltn)+1;
    
    % select the most promising future
    if(recon.tsnrmin(sminj) > recon.tsnrmin(smini))
        recon.tscore = recon.tscore2(smini,:);
        recon.t(ei) = recon.talt(sminj);
    else
        recon.tscore = recon.tscore2(:,sminj);
        recon.t(ei) = recon.talt(smini);
    end
    recon.t = recon.t(1:ei); % cleanup
    
    % store current estimation for next iteration
    [snrmux varmux mux tvarmuxS tmuxS twmuxS] = ...
        topt_evalvarx_full(cam, recon.X.mu, recon.t, false);
    varmuxS = tvarmuxS;  muxS = tmuxS;  wmuxS = twmuxS;
    recon.minsnr = min(snrmux(:));
    
    % print optimum
    if(recon.t(ei) > 1)
        tstr = sprintf('%d', recon.t(ei));
    else
        tstr = sprintf('1/%d', 1./recon.t(ei));
    end
    disp(sprintf('t(%2d)=%12s, min(SNR)=%12g, mean(SNR)=%12g', ei, tstr, ...
        20*log10(recon.minsnr), 20*log10(mean(snrmux(:)))));
        
    % plot search space
    if X_plot_process
        coli = mod(ei-1, 5)+1;
        
        figure(1)
        semilogx(recon.talt, recon.tscore, ...
            sprintf('%d-;N=%d;', coli, ei));
        
        % plot optimum
        text(recon.t(ei), smin, sprintf('%d', ei));
        
        % plot SNR
        figure(2), hold on
        semilogx(recon.X.mu(tsi), 20*log10(snrmux(tsi)), ...
            sprintf('-%d', coli));
    end    
    
    ei = ei + 1;    
end


% plot SNR and exposure time ranges
if X_plot_process
    
    tei = [3 8 15];
    tsnr=[15 25 30];
    
    figure(2),clf,hold on
    xlabel('Irradiance');
    ylabel('SNR (dB)');
    for i=1:length(tsnr)
        semilogx([1e2 1e6], ...
            [tsnr(i) tsnr(i)], sprintf('-%d',i));
    end
    
    recon.t = recon.t(1:15);
    [tt tti] = sort(recon.t);
    recon.nt = length(recon.t);
    for ei=tei
        [snrmux varmux mux tvarmuxS tmuxS twmuxS] = ...
            topt_evalvarx_full(cam, recon.X.mu, recon.t(1:ei), false);
        semilogx(recon.X.mu(tsi), 20*log10(snrmux(tsi)), ...
            sprintf('-%d;N=%2d;',find(ei == tei), ei));
        20*log10(min(snrmux(:)))
    end
    
    yf = 1.05;
    xf = 1.1;
    ymax = 54;
    ymin = 35.5;
    cm = jet(recon.nt);
    for ei=1:recon.nt
        if(recon.t(tti(ei)) > 0.5)
            tstr = sprintf('%.1g', recon.t(tti(ei)));
        else
            tstr = sprintf('1/%d', round(1./recon.t(tti(ei))));
        end
        colrgb = cm(ei,:);
        muxsat = (cam.vsat-cam.R.mu-1)./(cam.g*tt(ei));
        yoff = ymax - tti(ei)*yf;
        semilogx([muxsat muxsat], [yoff ymin], '-k'); % vertical
        line([minx muxsat], [yoff yoff], 'color', cm(tti(ei),:)); % horizontal
        text(muxsat*xf, yoff, sprintf('t_{%d}=%s',tti(ei),tstr)); % t label
    end
    legend('location', 'southeast')
    %print -debug=etopt_test2v2.plot etopt_test2v2.eps
    
    
    figure(3), clf
    hist(log10(recon.X.mu(:)),30)
    xlabel('Irradiance')
    %print -debug=etopt_test2v2_hist.plot etopt_test2v2_hist.eps
end